//! Subscription manager
use super::AllClients;
use crate::{client, Client, Error};
use chrono::Utc;
use futures::future::join_all;
use nostr_rs_types::{
    client::subscribe::{self},
    types::SubscriptionId,
    Response,
};
use std::{
    collections::{BTreeMap, VecDeque},
    sync::{
        atomic::{AtomicUsize, Ordering},
        Arc,
    },
};
use tokio::sync::{mpsc, RwLock, RwLockWriteGuard};
use url::Url;

#[derive(Debug, Copy, Default, Eq, PartialEq, Clone)]
/// Subscription status
pub enum Status {
    /// Subscription is awaiting to be subscribed
    #[default]
    Queued,
    /// Subscribed is active and it is fetching previous records and no EOD has
    /// been received
    Fetching,
    /// Subscription is listening, an EOD has been received. This state can be
    /// Requeued is their spot is needed for other subscriptions
    Subscribed,
    /// Waiting to be subscribed again
    Requeued,
    /// Resubscribed, like subscribed but the EOD is ignored and nore relayed to
    /// the listeners, since this is not the first the this subscription has
    /// been created and it will be rotated soon
    Resubscribed,
}

#[derive(Debug, Default)]
struct SubscriptionInner {
    /// Active subscription (in the client side), when this is Drop all clients unsubscribes
    active_subscription: Option<Vec<client::ActiveSubscription>>,
    /// Subscription status
    status: Status,
    /// raw request
    subscription_request: subscribe::Subscribe,
}

/// Active subscription
///
/// This object is responsible for keeping track of a subscription for a
/// connection
///
/// This must be dropped to unsubscribe from the subscription manager
pub struct ActiveSubscription {
    unsubscriber: Option<(PoolSubscriptionId, Arc<Scheduler>)>,
}

impl Drop for ActiveSubscription {
    fn drop(&mut self) {
        if let Some((id, scheduler)) = self.unsubscriber.take() {
            scheduler.remove(id);
        }
    }
}

/// Pool subscription ID
#[derive(Debug, Clone, Eq, PartialEq, Hash, Ord, PartialOrd)]
pub struct PoolSubscriptionId((SubscriptionId, Option<Url>));

impl Default for PoolSubscriptionId {
    fn default() -> Self {
        Self((SubscriptionId::empty(), None))
    }
}

/// Subscription manager
///
/// Pools can have an unlimited number of active subscriptions, but this
/// subscriptions will be scheduled internally.
#[derive(Default)]
pub(crate) struct Scheduler {
    subscriptions: RwLock<BTreeMap<PoolSubscriptionId, SubscriptionInner>>,
    all_clients: AllClients,
    /// VecDec to have the list of subscritions
    ///
    /// This structure will pop the first elements and move them to be back
    /// waiting their turn to be rescheduled.
    subscription_queue: RwLock<VecDeque<PoolSubscriptionId>>,
    active_subscriptions: Arc<AtomicUsize>,
    total_subscriptions: Arc<AtomicUsize>,
}

/// Maximum number of subscriptions
pub const MAX_SUBSCRIPTIONS: usize = 5;

impl Scheduler {
    /// Creates a new instance
    pub fn new(all_clients: AllClients) -> Self {
        Self {
            all_clients,
            ..Default::default()
        }
    }

    /// Register a new Client
    pub async fn register_client(&self, client: &Client) {
        let mut subscriptions = self.subscriptions.write().await;
        for subscription in subscriptions.values_mut() {
            if matches!(
                subscription.status,
                Status::Fetching | Status::Resubscribed | Status::Subscribed,
            ) {
                if let Ok(active_subscription) = client
                    .subscribe(subscription.subscription_request.clone())
                    .await
                {
                    if let Some(active_subscriptions) = subscription.active_subscription.as_mut() {
                        active_subscriptions.push(active_subscription);
                    }
                }
            }
        }
    }

    /// Processes all messages from the client pools
    ///
    /// Mainly to schedule subscriptions and pass the message over to the
    /// listeners
    pub async fn process_message(
        self: &Arc<Self>,
        message: Response,
        url: Url,
        return_to: mpsc::Sender<(Response, Url)>,
    ) -> Result<(), Error> {
        match message {
            Response::EndOfStoredEvents(subscription_id) => {
                let subscription_id = PoolSubscriptionId((subscription_id.0, None));
                let mut subscription = self.subscriptions.write().await;
                if let Some(s) = subscription.get_mut(&subscription_id) {
                    let old_status = s.status;
                    s.status = Status::Subscribed;

                    if old_status == Status::Fetching {
                        return_to
                            .try_send((
                                Response::EndOfStoredEvents(subscription_id.0 .0.into()),
                                url,
                            ))
                            .map_err(|e| Error::InternalChannel(e.to_string()))?;
                    }
                }

                self.active_subscription_scheduler();

                Ok(())
            }
            any_message => {
                return_to
                    .try_send((any_message, url))
                    .map_err(|e| Error::InternalChannel(e.to_string()))?;
                Ok(())
            }
        }
    }

    fn active_subscription_scheduler(self: &Arc<Self>) {
        let this = self.clone();
        tokio::spawn(async move {
            let clients = this.all_clients.read().await;
            let mut subscriptions = this.subscriptions.write().await;
            let mut subscription_queue = this.subscription_queue.write().await;
            let items = subscription_queue.len();

            // A subscription must be descheduled as its place is needed.
            let deschedule =
                |subscriptions: &mut RwLockWriteGuard<
                    '_,
                    BTreeMap<PoolSubscriptionId, SubscriptionInner>,
                >,
                 subscription_queue: &mut RwLockWriteGuard<'_, VecDeque<PoolSubscriptionId>>|
                 -> bool {
                    for subscription_id in subscription_queue.iter() {
                        let subscription =
                            if let Some(subscription) = subscriptions.get_mut(subscription_id) {
                                subscription
                            } else {
                                continue;
                            };

                        if matches!(subscription.status, Status::Subscribed) {
                            // unsubscribe
                            let _ = subscription.active_subscription.take();
                            // update counter
                            this.active_subscriptions.fetch_sub(1, Ordering::Relaxed);
                            // update since for next request
                            let now = Utc::now();
                            for filter in subscription.subscription_request.filters.iter_mut() {
                                filter.since = Some(now);
                            }
                            return true;
                        }
                    }

                    false
                };

            for _ in 0..items {
                let subscription_id = if let Some(subscription_id) = subscription_queue.pop_front()
                {
                    subscription_id
                } else {
                    continue;
                };
                let mut subscription =
                    if let Some(subscription) = subscriptions.remove(&subscription_id) {
                        subscription
                    } else {
                        continue;
                    };

                // add subscription id back to the last element, to be visited later
                subscription_queue.push_back(subscription_id.clone());

                let prev_status = subscription.status;

                if matches!(prev_status, Status::Queued | Status::Requeued) {
                    if this.active_subscriptions.load(Ordering::SeqCst) >= MAX_SUBSCRIPTIONS
                        && !deschedule(&mut subscriptions, &mut subscription_queue)
                    {
                        subscriptions.insert(subscription_id, subscription);
                        // There is no more room to promote this subscription to active
                        break;
                    }

                    let wait_all = clients
                        .values()
                        .map(|(_, sender)| {
                            sender.subscribe(subscription.subscription_request.clone())
                        })
                        .collect::<Vec<_>>();

                    if let Ok(active_subscriptions) = join_all(wait_all)
                        .await
                        .into_iter()
                        .collect::<Result<Vec<_>, _>>()
                    {
                        subscription.active_subscription = Some(active_subscriptions);
                        subscription.status = if prev_status == Status::Queued {
                            Status::Fetching
                        } else {
                            Status::Resubscribed
                        };

                        this.active_subscriptions.fetch_add(1, Ordering::Relaxed);
                    }
                }

                // add it back
                subscriptions.insert(subscription_id, subscription);
            }
        });
    }

    /// Creates a new subscription with a given filters
    pub async fn subscribe(
        self: &Arc<Self>,
        subscription_request: subscribe::Subscribe,
        specific_url: Option<Url>,
    ) -> ActiveSubscription {
        let subscription_id = PoolSubscriptionId((
            subscription_request.subscription_id.clone(),
            specific_url.clone(),
        ));

        self.subscriptions.write().await.insert(
            subscription_id.clone(),
            SubscriptionInner {
                status: Status::Queued,
                active_subscription: None,
                subscription_request,
            },
        );

        // add subscription to the queue, to be scheduled eventually
        self.subscription_queue
            .write()
            .await
            .push_back(subscription_id.clone());

        self.total_subscriptions.fetch_add(1, Ordering::Relaxed);
        self.active_subscription_scheduler();

        ActiveSubscription {
            unsubscriber: Some((subscription_id, self.clone())),
        }
    }

    fn remove(self: Arc<Self>, subscription_id: PoolSubscriptionId) {
        let this = self;
        tokio::spawn(async move {
            let mut subscriptions = this.subscriptions.write().await;
            if subscriptions.remove(&subscription_id).is_some() {
                this.active_subscription_scheduler();
                this.total_subscriptions.fetch_sub(1, Ordering::Relaxed);
            }
        });
    }

    /// Total number of subscribers
    pub fn total_subscribers(&self) -> usize {
        self.total_subscriptions.load(Ordering::Relaxed)
    }
}
